/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 *
 * EGA/VGA Screen Driver, PC bios version
 * 	This driver uses int10 bios to set/reset graphics/text modes,
 * 	and to get the address of the ROM character font which
 * 	is used for the character bitmaps.  All other access to the
 * 	hardware is controlled through this driver.
 *
 * 	This driver links with one of two other files, vgaplan4.c,
 * 	the portable VGA 4 planes 16 color driver, or asmplan4.s, which
 * 	is 8086 assembly language for speed.  This file itself
 * 	doesn't know about any planar or packed arrangement, relying soley
 * 	on the following external routines for all graphics drawing:
 * 		ega_init, ega_drawpixel, ega_readpixel,
 * 		ega_drawhline, ega_drawvline
 *
 * 	All text/font drawing code is based on the above routines and
 * 	the included entry points for getting the ROM bitmap data.  Compiled
 * 	in fonts aren't supported for size reasons.  scr_bogl supports them.
 *
 * 	If the environment variable EGAMODE is set, the driver implements
 *	the EGA 640x350 (mode 10h) resolution, otherwise 640x480 (mode 12h)
 *	graphics mode is set.
 *
 *	The environment variable CHARHEIGHT if set will set the assumed rom
 *	font character height, which defaults to 14.
 *
 * 	If #define HWINIT 1 is set, the file vgainit.c is
 * 	used to provide non-bios direct hw initialization of the VGA
 * 	chipset.
 */
#define HWINIT	0		/* =1 for non-bios direct hardware init*/

#if ELKS
#include <linuxmt/ntty.h>
#endif
#include <stdlib.h>
#include "../device.h"
#include "vgaplan4.h"

/* VGA driver entry points*/
static int  VGA_open(SCREENDEVICE *psd);
static void VGA_close(void);
static void VGA_getscreeninfo(PSCREENINFO psi);;
static void VGA_setpalette(int first,int count,RGBENTRY *pal);
static void VGA_drawpixel(COORD x, COORD y, PIXELVAL c);
static PIXELVAL VGA_readpixel(COORD x, COORD y);
static void VGA_drawhline(COORD x1, COORD x2, COORD y, PIXELVAL c);
static void VGA_drawvline(COORD x, COORD y1, COORD y2, PIXELVAL c);
static void VGA_fillrect(COORD x1, COORD y1, COORD x2, COORD y2, PIXELVAL c);

/* VGA driver bios ROM font routines*/
static BOOL VGA_getfontinfo(FONTID fontid,PFONTINFO pfontinfo);
static void VGA_gettextsize(const UCHAR *str,int cc,COORD *retwd,
		COORD *retht,FONTID fontid);
static void VGA_gettextbits(UCHAR ch,IMAGEBITS *retmap,COORD *retwd,
		COORD *retht,FONTID fontid);
/*static void VGA_drawtext(COORD x, COORD y, const UCHAR *s, int n, PIXELVAL fg,
		FONTID fontid);
static void gen_drawbitmap(COORD x, COORD y, COORD width, COORD height,
			IMAGEBITS *table, PIXELVAL fgcolor);*/

SCREENDEVICE	scrdev = {
	VGA_open,
	VGA_close,
	VGA_getscreeninfo,
	VGA_setpalette,
	VGA_drawpixel,
	VGA_readpixel,
	VGA_drawhline,
	VGA_drawvline,
	VGA_fillrect,
	VGA_getfontinfo,
	VGA_gettextsize,
	VGA_gettextbits,
	640, 480, 256
};

/* compiled in fonts*/
#define NUMBER_FONTS	1	/* rom font only for now*/

/* operating mode*/
BOOL VGAMODE = TRUE;		/* ega or vga screen rows*/

/* rom font constants*/
int	ROM_CHAR_HEIGHT = 14;	/* number of scan lines in fonts in ROM */
#define	ROM_CHAR_WIDTH	8	/* number of pixels for character width */
#define MAX_ROM_HEIGHT	16	/* max rom character height*/
#define	FONT_CHARS	256	/* number of characters in font tables */

/* int10 functions*/
#define FNGR640x480	0x0012	/* function for graphics mode 640x480x16*/
#define FNGR640x350	0x0010	/* function for graphics mode 640x350x16*/
#define FNTEXT		0x0003	/* function for 80x25 text mode*/
#define FNGETROMADDR	0x1130	/* function for address of rom character table*/
#define GETROM8x14	0x0200	/* want address of ROM 8x14 char table*/

/* local data*/
FARADDR 		rom_char_addr;

/* external routines implementing planar ega/vga access*/

/* vgaplan4.c portable C driver, or asmplan4.s asm driver*/
void		ega_init(void);
void 		ega_drawpixel(unsigned int x,unsigned int y,int c);
PIXELVAL 	ega_readpixel(unsigned int x,unsigned int y);

/* vgaplan4 driver only*/
void		ega_drawhline(unsigned int x1,unsigned int x2,unsigned int y,int c);
void		ega_drawvline(unsigned int x,unsigned int y1,unsigned int y2,int c);

/* asmplan4 driver only*/
void 		ega_drawline(unsigned int x1,unsigned int y1,unsigned int x2,unsigned int y2,int c);

/* vgainit.c - direct hw init*/
void		ega_hwinit(void);
void		ega_hwterm(void);


static int
VGA_open(SCREENDEVICE *psd)
{
	char *p;

	/* setup operating mode from environment variable*/
	if(getenv("EGAMODE"))
		VGAMODE = FALSE;
	else VGAMODE = TRUE;

#if ELKS
	/* disallow console switching while in graphics mode*/
	if(ioctl(0, DCGET_GRAPH) != 0)
		return -1;
#endif

	/* enter graphics mode*/
#if HWINIT
	ega_hwinit();
#else
	/* init bios graphics mode*/
	int10(VGAMODE? FNGR640x480: FNGR640x350, 0);
#endif

	/* get address of rom character table*/
	rom_char_addr = int10(FNGETROMADDR, GETROM8x14);

#if 0
	/* check bios data area for actual character height,
	 * as the returned font isn't always 14 high
	 */
	ROM_CHAR_HEIGHT = GETBYTE_FP(MK_FP(0x0040, 0x0085));
	if(ROM_CHAR_HEIGHT > MAX_ROM_HEIGHT)
		ROM_CHAR_HEIGHT = MAX_ROM_HEIGHT;
#endif

	/* init driver variables depending on ega/vga mode*/
	psd->xres = 640;
	psd->yres = VGAMODE? 480: 350;
	psd->ncolors = 16;
#if 0
	ROM_CHAR_HEIGHT = VGAMODE? 16: 14;
#endif
	p = getenv("CHARHEIGHT");
	if(p)
		ROM_CHAR_HEIGHT = atoi(p);


	/* init planes driver*/
	ega_init();

	/* FIXME: add palette code*/
	return 1;
}

static void
VGA_close(void)
{
#if ELKS
	/* allow console switching again*/
	ioctl(0, DCREL_GRAPH);
#endif
#if HWINIT
	ega_hwterm();
#else
	/* init bios 80x25 text mode*/
	int10(FNTEXT, 0);
#endif
}

static void
VGA_getscreeninfo(PSCREENINFO psi)
{
	psi->rows = scrdev.yres;
	psi->cols = scrdev.xres;
	psi->ncolors = scrdev.ncolors;
	psi->black = 0;
	psi->white = 15;
	psi->fonts = NUMBER_FONTS;

	if(VGAMODE) {
		/* VGA 640x480*/
		psi->xdpcm = 27;		/* assumes screen width of 24 cm*/
		psi->ydpcm = 27;		/* assumes screen height of 18 cm*/
	} else {
		/* EGA 640x350*/
		psi->xdpcm = 27;		/* assumes screen width of 24 cm*/
		psi->ydpcm = 19;		/* assumes screen height of 18 cm*/
	}

#if ETA4000
	/* SVGA 800x600*/
	psi->xdpcm = 33;		/* assumes screen width of 24 cm*/
	psi->ydpcm = 33;		/* assumes screen height of 18 cm*/
#endif
}

static void
VGA_setpalette(int first,int count,RGBENTRY *pal)
{
	/* not yet implemented, std 16 color palette assumed*/
}

static void
VGA_drawpixel(COORD x, COORD y, PIXELVAL c)
{
	ega_drawpixel(x, y, c);
}

static PIXELVAL
VGA_readpixel(COORD x, COORD y)
{
	return ega_readpixel(x, y);
}

static void
VGA_drawhline(COORD x1, COORD x2, COORD y, PIXELVAL c)
{
	ega_drawhline(x1, x2, y, c);

	/*ega_drawline(x1, y, x2, y, c);*/
	/*
	 * Uncomment the following if driver doesn't support hline
	while(x1 <= x2)
		VGA_drawpixel(x1++, y, c);
	 */
}

static void
VGA_drawvline(COORD x, COORD y1, COORD y2, PIXELVAL c)
{
	ega_drawvline(x, y1, y2, c);

	/*ega_drawline(x, y1, x, y2, c);*/
	/*
	 * Uncomment the following if driver doesn't support vline
	while(y1 <= y2)
		VGA_drawpixel(x, y1++, c);
	 */
}

static void
VGA_fillrect(COORD x1, COORD y1, COORD x2, COORD y2, PIXELVAL c)
{
	while(y1 <= y2)
		VGA_drawhline(x1, x2, y1++, c);
}

#if 0000
/* 
 * Low level text draw routine, called only if no clipping
 * is required.  This routine draws ROM font characters only.
 */
static void
VGA_drawtext(COORD x,COORD y,const UCHAR *s,int n,PIXELVAL fg,FONTID fontid)
{
	COORD 		width;			/* width of character */
	COORD 		height;			/* height of character */
	IMAGEBITS 	bitmap[MAX_ROM_HEIGHT];	/* bitmap for character */

 	/* x,y is bottom left corner*/
	y -= (ROM_CHAR_HEIGHT - 1);
	while (n-- > 0) {
		VGA_gettextbits(*s++, bitmap, &width, &height, pfont);
		gen_drawbitmap(x, y, width, height, bitmap, fg);
		x += width;
	}
}

/*
 * Generalized low level bitmap output routine, called
 * only if no clipping is required.  Only the set bits
 * in the bitmap are drawn, in the foreground color.
 */
static void
gen_drawbitmap(COORD x, COORD y, COORD width, COORD height, IMAGEBITS *table,
	PIXELVAL fgcolor)
{
  COORD minx;
  COORD maxx;
  IMAGEBITS bitvalue;	/* bitmap word value */
  int bitcount;			/* number of bits left in bitmap word */

  minx = x;
  maxx = x + width - 1;
  bitcount = 0;
  while (height > 0) {
	if (bitcount <= 0) {
		bitcount = IMAGE_BITSPERIMAGE;
		bitvalue = *table++;
	}
	if (IMAGE_TESTBIT(bitvalue))
		ega_drawpixel(x, y, fgcolor);
	bitvalue = IMAGE_SHIFTBIT(bitvalue);
	--bitcount;
	if (x++ == maxx) {
		x = minx;
		++y;
		--height;
		bitcount = 0;
	}
  }
}
#endif

/*
 * VGA low level get font info routine.  This routine
 * returns info on a single bios ROM font.
 */
static BOOL
VGA_getfontinfo(FONTID fontid,PFONTINFO pfontinfo)
{
	int	i;

	pfontinfo->font = 0;
	pfontinfo->height = ROM_CHAR_HEIGHT;
	pfontinfo->maxwidth = ROM_CHAR_WIDTH;
	pfontinfo->baseline = 3;
	pfontinfo->fixed = TRUE;
	for (i = 0; i < 256; i++)
		pfontinfo->widths[i] = ROM_CHAR_WIDTH;
	return TRUE;
}

/*
 * VGA low level routine to calc bounding box for text output.
 * Handles bios ROM font only.
 */
static void
VGA_gettextsize(const UCHAR *str,int cc,COORD *retwd,COORD *retht,
	FONTID fontid)
{
	*retwd = ROM_CHAR_WIDTH * cc;
	*retht = ROM_CHAR_HEIGHT;
}

/*
 * VGA low level routine to get the bitmap associated
 * with a character.  Handles bios ROM font only.
 */
static void
VGA_gettextbits(UCHAR ch,IMAGEBITS *retmap,COORD *retwd, COORD *retht,
	FONTID fontid)
{
	FARADDR	bits;
	int	n;

	/* read character bits from rom*/
	bits = rom_char_addr + ch * ROM_CHAR_HEIGHT;
	for(n=0; n<ROM_CHAR_HEIGHT; ++n)
		*retmap++ = GETBYTE_FP(bits++) << 8;

	*retwd = ROM_CHAR_WIDTH;
	*retht = ROM_CHAR_HEIGHT;
}
