/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 *
 * Screen Driver using SVGA Library
 *
 * This driver requires the following SVGA entry points:
 * 	vga_init, vga_setmode,
 * 	vga_drawpixel, vga_getpixel,
 * 	vga_setegacolor, vga_drawline,
 *	vga_getscansegment, vga_drawscansegment
 *
 * All graphics drawing primitives are based on top of these functions.
 *
 * This file also contains the generalized low-level font/text
 * drawing routines, which will be split out into another file.
 * Both fixed and proportional fonts are supported.
 */
#include <stdio.h>
#include "../device.h"
#include <vga.h>

/* specific vgalib driver entry points*/
static int  SVGA_open(SCREENDEVICE *psd);
static void SVGA_close(void);
static void SVGA_getscreeninfo(PSCREENINFO psi);
static void SVGA_setpalette(int first,int count,RGBENTRY *pal);
static void SVGA_drawpixel(COORD x, COORD y, PIXELVAL c);
static PIXELVAL SVGA_readpixel(COORD x, COORD y);
static void SVGA_drawhline(COORD x1, COORD x2, COORD y, PIXELVAL c);
static void SVGA_drawvline(COORD x, COORD y1, COORD y2, PIXELVAL c);
static void SVGA_fillrect(COORD x1, COORD y1, COORD x2, COORD y2, PIXELVAL c);

/* generalized text/font routines*/
static BOOL gen_getfontinfo(FONTID fontid,PFONTINFO pfontinfo);
static void gen_gettextsize(const UCHAR *str,int cc,COORD *retwd,
		COORD *retht,FONTID fontid);
static void gen_gettextbits(UCHAR ch,IMAGEBITS *retmap,COORD *retwd,
		COORD *retht,FONTID fontid);
/*static void gen_drawtext(COORD x, COORD y, const UCHAR *s, int n,
			PIXELVAL fg, FONTID fontid);
static void gen_drawbitmap(COORD x, COORD y, COORD width, COORD height,
			IMAGEBITS *table, PIXELVAL fgcolor);*/

SCREENDEVICE	scrdev = {
	SVGA_open,
	SVGA_close,
	SVGA_getscreeninfo,
	SVGA_setpalette,
	SVGA_drawpixel,
	SVGA_readpixel,
	SVGA_drawhline,
	SVGA_drawvline,
	SVGA_fillrect,
	gen_getfontinfo,
	gen_gettextsize,
	gen_gettextbits,
	640, 480, 256
};

/* compiled in fonts*/
#define NUMBER_FONTS	3
extern FONT font_rom8x16, font_rom8x8;
extern FONT font_winFreeSansSerif11x13;
extern FONT font_winFreeSystem14x16;
extern FONT font_winSystem14x16;
extern FONT font_winMSSansSerif11x13;
extern FONT font_winTerminal8x12;
extern FONT font_helvB10, font_helvB12, font_helvR10;

/* first font is default font*/
PFONT fonts[NUMBER_FONTS] = {
#if HAVEMSFONTS
	&font_winSystem14x16,		/* FONT_SYSTEM_VAR*/
	&font_winMSSansSerif11x13,	/* FONT_GUI_VAR*/
	&font_winTerminal8x12		/* FONT_OEM_FIXED*/
#else
	&font_winFreeSystem14x16,	/* FONT_SYSTEM_VAR*/
	&font_winFreeSansSerif11x13,	/* FONT_GUI_VAR*/
	&font_rom8x16			/* FONT_OEM_FIXED*/
#endif
};

extern MODE gr_mode;	/* temp kluge*/

static int
SVGA_open(SCREENDEVICE *psd)
{
	int		mode;
	vga_modeinfo *	modeinfo;

	vga_init();
	//mode = G640x480x256;
	mode = G640x480x16;
	vga_setmode(mode);
	modeinfo = vga_getmodeinfo(mode);

	psd->xres = modeinfo->width;
	psd->yres = modeinfo->height;
	psd->ncolors = modeinfo->colors;
	return 1;
}

static void
SVGA_close(void)
{
	vga_setmode(TEXT);
}

static void
SVGA_getscreeninfo(PSCREENINFO psi)
{
	psi->rows = scrdev.yres;
	psi->cols = scrdev.xres;
	psi->ncolors = scrdev.ncolors;
	psi->black = 0;
	psi->white = 15;
	psi->fonts = NUMBER_FONTS;

	if(scrdev.yres > 480) {
		/* SVGA 800x600*/
		psi->xdpcm = 33;	/* assumes screen width of 24 cm*/
		psi->ydpcm = 33;	/* assumes screen height of 18 cm*/
	} else if(scrdev.yres > 350) {
		/* VGA 640x480*/
		psi->xdpcm = 27;	/* assumes screen width of 24 cm*/
		psi->ydpcm = 27;	/* assumes screen height of 18 cm*/
	} else {
		/* EGA 640x350*/
		psi->xdpcm = 27;	/* assumes screen width of 24 cm*/
		psi->ydpcm = 19;	/* assumes screen height of 18 cm*/
	}
}

static void
SVGA_setpalette(int first,int count,RGBENTRY *pal)
{
	while(first < 256 && count-- > 0) {
		vga_setpalette(first++, pal->r>>2, pal->g>>2, pal->b>>2);
		++pal;
	}
}

static void
SVGA_drawpixel(COORD x, COORD y, PIXELVAL c)
{
	unsigned char gline, line;

	if(gr_mode == MODE_SET) {
		vga_setegacolor(c);
		vga_drawpixel(x, y);
		return;
	}
	/*
	 * This fishery is required because vgalib doesn't support
	 * xor drawing mode without acceleration.
	 */
	vga_getscansegment(&gline, x, y, 1);
	line = c;
	line ^= gline;
	vga_drawscansegment(&line, x, y, 1);
}

static PIXELVAL
SVGA_readpixel(COORD x, COORD y)
{
	return vga_getpixel(x, y);
}

static void
SVGA_drawhline(COORD x1, COORD x2, COORD y, PIXELVAL c)
{
	int 	i, width;
	unsigned char getline[640];
	static int lastcolor = -1;
	static int lastwidth = -1;
	static unsigned char line[640];

	/*
	 * All this fishery is required for two reasons:
	 * one, vga_drawline is way too slow to be called to fill
	 * rectangles, so vga_drawscansegment is used instead.  In
	 * addition, vgalib doesn't support xor drawing mode
	 * without acceleration!!, so we've got to do it ourselves
	 * with vga_getscansegment...
	 */
	width = x2-x1+1;

	/* this is faster than calling vga_drawline !!!*/
	if(width != lastwidth || c != lastcolor) {
		lastwidth = width;
		lastcolor = c;
		for(i=0; i<width; ++i)
			line[i] = c;
	}
	if(gr_mode == MODE_XOR) {
		vga_getscansegment(getline, x1, y, width);
		for(i=0; i<width; ++i)
			line[i] ^= getline[i];
		lastwidth = -1;
	}
	vga_drawscansegment(line, x1, y, width);

	/*
	 * Non-fishery version is *slow* and doesn't support XOR.
	vga_setegacolor(c);
	vga_drawline(x1, y, x2, y2);
	 */
}

static void
SVGA_drawvline(COORD x, COORD y1, COORD y2, PIXELVAL c)
{
	if(gr_mode == MODE_SET) {
		vga_setegacolor(c);
		vga_drawline(x, y1, x, y2);
	}

	/* slower version required for XOR drawing support*/
	while(y1 <= y2)
		SVGA_drawpixel(x, y1++, c);
}

static void
SVGA_fillrect(COORD x1, COORD y1, COORD x2, COORD y2, PIXELVAL c)
{
	while(y1 <= y2)
		SVGA_drawhline(x1, x2, y1++, c);
}

#if 0000
/* 
 * Generalized low level text draw routine, called only
 * if no clipping is required
 */
static void
gen_drawtext(COORD x,COORD y,const UCHAR *s,int n,PIXELVAL fg,FONTID fontid)
{
	COORD 		width;			/* width of character */
	COORD 		height;			/* height of character */
	PFONT		pf;
	IMAGEBITS 	bitmap[MAX_CHAR_HEIGHT];/* bitmap for character */

	if(fontid >= NUMBER_FONTS)
		return;
	pf = fonts[fontid];

	/* x, y is bottom left corner*/
	y -= pf->height - 1;
	while (n-- > 0) {
		gen_gettextbits(*s++, bitmap, &width, &height, pf);
		gen_drawbitmap(x, y, width, height, bitmap, fg);
		x += width;
	}
}

/*
 * Generalized low level bitmap output routine, called
 * only if no clipping is required.  Only the set bits
 * in the bitmap are drawn, in the foreground color.
 */
static void
gen_drawbitmap(COORD x, COORD y, COORD width, COORD height, IMAGEBITS *table,
	PIXELVAL fgcolor)
{
  COORD minx;
  COORD maxx;
  IMAGEBITS bitvalue;	/* bitmap word value */
  int bitcount;			/* number of bits left in bitmap word */

  minx = x;
  maxx = x + width - 1;
  bitcount = 0;
  while (height > 0) {
	if (bitcount <= 0) {
		bitcount = IMAGE_BITSPERIMAGE;
		bitvalue = *table++;
	}
	if (IMAGE_TESTBIT(bitvalue))
		SVGA_drawpixel(x, y, fgcolor);
	bitvalue = IMAGE_SHIFTBIT(bitvalue);
	--bitcount;
	if (x++ == maxx) {
		x = minx;
		++y;
		--height;
		bitcount = 0;
	}
  }
}
#endif

/*
 * Generalized low level get font info routine.  This
 * routine works with fixed and proportional fonts.
 */
static BOOL
gen_getfontinfo(FONTID fontid,PFONTINFO pfontinfo)
{
	PFONT	pf;
	int	i;

	if(fontid >= NUMBER_FONTS)
		return FALSE;
	pf = fonts[fontid];

	pfontinfo->font = fontid;
	pfontinfo->height = pf->height;
	pfontinfo->maxwidth = pf->maxwidth;
	pfontinfo->baseline = 0;
	pfontinfo->fixed = pf->width == NULL? TRUE: FALSE;
	for(i=0; i<256; ++i) {
		if(pf->width == NULL)
			pfontinfo->widths[i] = pf->maxwidth;
		else {
			if(i<pf->firstchar || i >= pf->firstchar+pf->size)
				pfontinfo->widths[i] = 0;
			else pfontinfo->widths[i] = pf->width[i-pf->firstchar];
		}
	}
	return TRUE;
}

/*
 * Generalized low level routine to calc bounding box for text output.
 * Handles both fixed and proportional fonts.
 */
static void
gen_gettextsize(const UCHAR *str,int cc,COORD *retwd,COORD *retht,
	FONTID fontid)
{
	PFONT	pf;
	int	c;
	int	width;

	if(fontid >= NUMBER_FONTS) {
		*retht = 0;
		*retwd = 0;
		return;
	}
	pf = fonts[fontid];

	if(pf->width == NULL)
		width = cc * pf->maxwidth;
	else {
		width = 0;
		while(--cc >= 0) {
			if( (c = *str++) >= pf->firstchar &&
			     c < pf->firstchar+pf->size)
				width += pf->width[c - pf->firstchar];
		}
	}

	*retwd = width;
	*retht = pf->height;

}

/*
 * Generalized low level routine to get the bitmap associated
 * with a character.  Handles fixed and proportional fonts.
 */
static void
gen_gettextbits(UCHAR ch,IMAGEBITS *retmap,COORD *retwd, COORD *retht,
	FONTID fontid)
{
	int		n;
	PFONT		pf = NULL;
	IMAGEBITS *	bits;

	if(fontid < NUMBER_FONTS)
		pf = fonts[fontid];

	if(!pf || ch < pf->firstchar || ch >= pf->firstchar+pf->size) {
		*retht = 0;
		*retwd = 0;
		return;
	}
	ch -= pf->firstchar;

	/* get font bitmap depending on fixed pitch or not*/
	bits = pf->bits + (pf->offset? pf->offset[ch]: (pf->height * ch));
	for(n=0; n<pf->height; ++n)
		*retmap++ = *bits++;

	/* return width depending on fixed pitch or not*/
	*retwd = pf->width? pf->width[ch]: pf->maxwidth;
	*retht = pf->height;
}
